from __future__ import annotations

from collections import namedtuple
from functools import partial
from itertools import chain, product, starmap, zip_longest
from operator import attrgetter, eq, is_not
from pathlib import Path
from typing import cast

from attr import Factory, attrib, attrs
from gherkin.pickles.compiler import Compiler
from marshmallow import Schema, fields, post_load, pre_load
from marshmallow_polyfield import PolyField

from pytest_bdd.ast import ASTSchema
from pytest_bdd.model import Feature
from pytest_bdd.utils import ModelSchemaPostLoadable, deepattrgetter, get_caller_module_locals, get_caller_module_path


class CastableToStrField(fields.Str):
    def __init__(
        self,
        *args,
        strict=False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.strict = strict

    def _deserialize(self, value, attr, data, **kwargs):
        if not self.strict and not isinstance(value, (str, bytes)):
            value = str(value)
        return super()._deserialize(value, attr, data, **kwargs)


@attrs
class Node:
    tags: list[str] = attrib(default=Factory(list))
    name: str | None = attrib(default=None)
    description: str | None = attrib(default=None)
    comments: list[str] = attrib(default=Factory(list))


class NodeSchema(Schema):
    tags = fields.List(fields.Str(), required=False, data_key="Tags")
    name = fields.Str(required=False, data_key="Name")
    description = fields.Str(required=False, data_key="Description")
    comments = fields.List(fields.Str(), required=False, data_key="Comments")


@attrs
class Step(Node, ModelSchemaPostLoadable):
    steps: list[Step] = attrib(default=Factory(list))
    action: str | None = attrib(default=None)
    type: str | None = attrib(default=None)

    data: list[Table | Join] = attrib(default=Factory(list))
    examples: list[Table | Join] = attrib(default=Factory(list))

    Route = namedtuple("Route", ["tags", "steps", "example_table"])

    @property
    def routes(self):
        for routes in (
            product(*map(attrgetter("routes"), self.steps))
            if self.steps
            else [[self.Route([], [], Table(parameters=[], values=[]))]]
        ):
            steps = [self, *chain.from_iterable(map(attrgetter("steps"), routes))]

            if self.examples:
                for _example_table in self.examples:
                    example_table = Join(tables=[*map(attrgetter("example_table"), routes), _example_table])
                    tags = list(
                        {*chain.from_iterable(map(attrgetter("tags"), routes)), *example_table.tags, *self.tags}
                    )

                    yield self.Route(
                        tags,
                        steps,
                        example_table,
                    )
            else:
                example_table = Join(tables=[*map(attrgetter("example_table"), routes)])
                tags = list({*chain.from_iterable(map(attrgetter("tags"), routes)), *example_table.tags, *self.tags})

                yield self.Route(
                    tags,
                    steps,
                    example_table,
                )

    def build_feature(self, filename, uri):
        from pytest_bdd.struct_bdd.ast_builder import DocumentASTBuilder

        gherkin_ast = DocumentASTBuilder(self).build()
        gherkin_ast.gherkin_document.uri = uri
        gherkin_ast_data = ASTSchema().dump(gherkin_ast)

        scenarios_data = Compiler().compile(gherkin_ast_data["gherkinDocument"])
        scenarios = Feature.load_scenarios(scenarios_data)

        instance = Feature(  # type: ignore[call-arg]
            gherkin_ast=gherkin_ast,
            uri=uri,
            scenarios=scenarios,
            filename=filename,
        )

        for scenario in scenarios:
            scenario.bind_feature(instance)

        return instance

    def build_resolver(self, filename, uri):
        feature = self.build_feature(filename=filename, uri=uri)

        class Resolver:
            @staticmethod
            def resolve():
                return zip_longest((), feature.scenarios, fillvalue=feature)

        return Resolver

    def _inject_test(
        self, caller_module_locals=None, caller_module_path=None, algorithm="build_bound_locator_test_function"
    ):
        from pytest_bdd.scenario import ModuleScenarioRegistry

        caller_locals = caller_module_locals or get_caller_module_locals(depth=3)
        caller_module_path = caller_module_path or get_caller_module_path(depth=3)

        method = getattr(
            ModuleScenarioRegistry.get(caller_locals=caller_locals, caller_module_path=caller_module_path), algorithm
        )
        return method(
            self.build_resolver(
                str(Path(caller_module_path).as_posix()), str(Path(caller_module_path).relative_to(Path.cwd()))
            )
        )

    def inject_test(self, caller_module_locals=None, caller_module_path=None):
        return self._inject_test(
            caller_module_locals=caller_module_locals,
            caller_module_path=caller_module_path,
            algorithm="build_bound_locator_test_function",
        )

    def build_test_decorator(self, caller_module_locals=None, caller_module_path=None):
        return self._inject_test(
            caller_module_locals=caller_module_locals,
            caller_module_path=caller_module_path,
            algorithm="build_bound_locator_test_decorator",
        )

    @staticmethod
    def from_dict(_dict) -> Step:
        return cast(Step, StepSchema().load(_dict))


class Given(Step):
    def __new__(cls, action, *args, **kwargs):
        return Step(*args, type="Given", action=action, **kwargs)


class When(Step):
    def __new__(cls, action, *args, **kwargs):
        return Step(*args, type="When", action=action, **kwargs)


class Then(Step):
    def __new__(cls, action, *args, **kwargs):
        return Step(*args, type="Then", action=action, **kwargs)


class Do(Step):
    def __new__(cls, action, *args, **kwargs):
        return Step(*args, action=action, **kwargs)


And = But = Do


KEYWORDS = ["given", "when", "then", "and", "but", "*"]


def step_deserialization_schema_selector(obj, parent_obj):
    if isinstance(obj, str):
        return StepStringSchema()

    try:
        key, *others = iter(obj.keys())
        if others:
            raise TypeError
        key = str(key).strip().lower()
    except StopIteration as e:
        raise TypeError from e

    if key in KEYWORDS:
        return StepKeywordSchema()
    elif key == "alternative":
        return AlternativeSchema()
    else:
        return StepContainerSchema()


def table_deserialization_schema_selector(obj, parent_obj):
    try:
        key, *others = iter(obj.keys())
        if others:
            raise TypeError
        key = str(key).strip().lower()
    except StopIteration as e:
        raise TypeError from e

    if key == "table":
        return TableContainerSchema()
    elif key == "join":
        return JoinSchema()
    else:
        raise TypeError


TablePolyfield = partial(
    PolyField,
    deserialization_schema_selector=table_deserialization_schema_selector,
)

StepPolyfield = partial(
    PolyField,
    deserialization_schema_selector=step_deserialization_schema_selector,
)


class StepSchema(NodeSchema):
    steps = StepPolyfield(many=True, data_key="Steps", required=False)

    action = fields.Str(required=False, data_key="Action")
    type = fields.Str(required=False, data_key="Type")
    data = TablePolyfield(many=True, data_key="Data", required=False)
    examples = TablePolyfield(many=True, data_key="Examples", required=False)

    schema_post_loader = Step.schema_post_loader()


class StepContainerSchema(Schema):
    step = fields.Nested(StepSchema, data_key="Step")

    @post_load
    def _(self, obj, *args, **kwargs):
        return obj["step"]


class StepStringSchema(StepSchema, Schema):
    @pre_load
    def convert(self, data, many, **kwargs):
        return {"Action": data}


class StepKeywordSchema(StepSchema, Schema):
    @pre_load
    def convert(self, data, many, **kwargs):
        return (lambda keyword, action: {"Action": action, "Type": keyword})(*next(iter(data.items())))


@attrs
class Alternative(ModelSchemaPostLoadable):
    steps: list[Step | Alternative] = attrib(default=Factory(list))

    @property
    def routes(self):
        yield from chain.from_iterable(map(attrgetter("routes"), self.steps))


class AlternativeSchema(Schema):
    steps = StepPolyfield(many=True, data_key="Alternative")

    schema_post_loader = Alternative.schema_post_loader()


@attrs
class Table(Node, ModelSchemaPostLoadable):
    type: str | None = attrib(default="Rowed")
    parameters: list[str] = attrib(default=Factory(list))
    values: list[list[str]] = attrib(default=Factory(list))

    @property
    def columned_values(self):
        if self.type == "Columned":
            return self.values
        else:
            return list(zip(*self.values))

    @property
    def rowed_values(self):
        if self.type == "Rowed":
            return self.values
        else:
            return list(zip(*self.values))


class TableSchema(NodeSchema):
    type = fields.Str(required=False, data_key="Type")

    parameters = fields.List(
        CastableToStrField(),
        data_key="Parameters",
    )
    values = fields.List(fields.List(CastableToStrField()), data_key="Values")

    schema_post_loader = Table.schema_post_loader()


class TableContainerSchema(Schema):
    table = fields.Nested(lambda: TableSchema(), data_key="Table")

    @post_load
    def _(self, obj, *args, **kwargs):
        return obj["table"]


@attrs
class Join(ModelSchemaPostLoadable):
    tables: list[Table] = attrib(default=Factory(list))

    __hash__ = id

    @property
    def tags(self):
        return list({*chain.from_iterable(map(attrgetter("tags"), self.tables))})

    @property
    def name(self):
        return "\n".join(filter(partial(is_not, None), reversed(list(chain(map(attrgetter("name"), self.tables))))))

    @property
    def description(self):
        return "\n".join(
            filter(
                partial(is_not, None),
                chain.from_iterable(map(deepattrgetter("description", skip_missing=True), self.tables)),
            )
        )

    @property
    def comments(self):
        return list(chain(map(attrgetter("comments"), self.tables)))

    @property
    def parameters(self):
        return list(set(chain.from_iterable(map(attrgetter("parameters"), self.tables))))

    @property
    def type(self):
        return "Rowed"

    @property
    def values(self):
        def _():
            filled_tables = list(filter(attrgetter("parameters"), self.tables))
            if filled_tables:
                filled_tables_parameters = list(chain.from_iterable(map(attrgetter("parameters"), self.tables)))
                for filled_tables_values in map(
                    lambda tables_values: list(chain.from_iterable(tables_values)),
                    product(*map(attrgetter("rowed_values"), filled_tables)),
                ):
                    if all(
                        [
                            all(
                                starmap(
                                    eq,
                                    product(
                                        [
                                            value
                                            for _parameter, value in zip(filled_tables_parameters, filled_tables_values)
                                            if parameter == _parameter
                                        ],
                                        repeat=2,
                                    ),
                                )
                            )
                            for parameter in self.parameters
                        ]
                    ):

                        def values_gen():
                            for parameter in self.parameters:
                                for _parameter, value in zip(filled_tables_parameters, filled_tables_values):
                                    if parameter == _parameter:
                                        yield value
                                        break

                        values = list(values_gen())
                        yield values
            else:
                yield from map(
                    lambda values_combination: list(chain.from_iterable(values_combination)),
                    product(*map(attrgetter("rowed_values"), self.tables)),
                )

        _values = list(_())
        return _values

    @property
    def columned_values(self):
        return list(zip(*self.values))

    @property
    def rowed_values(self):
        return self.values


class JoinSchema(Schema):
    tables = TablePolyfield(many=True, data_key="Join")

    schema_post_loader = Join.schema_post_loader()



"""Various utility functions."""
from __future__ import annotations

import base64
import pickle
import re
from collections import defaultdict
from contextlib import nullcontext, suppress
from functools import partial
from inspect import getframeinfo, signature
from itertools import tee
from operator import attrgetter, getitem, itemgetter
from sys import _getframe
from typing import TYPE_CHECKING, Any, Callable, Collection, Mapping, cast

from attr import Factory, attrib, attrs
from marshmallow import post_load

from pytest_bdd.const import ALPHA_REGEX, PYTHON_REPLACE_REGEX
from pytest_bdd.typing.pytest import FixtureDef

if TYPE_CHECKING:  # pragma: no cover
    from pytest_bdd.typing.pytest import RunResult


def get_args(func: Callable) -> list[str]:
    """Get a list of argument names for a function.

    :param func: The function to inspect.

    :return: A list of argument names.
    :rtype: list
    """
    params = signature(func).parameters.values()
    return [param.name for param in params if param.kind == param.POSITIONAL_OR_KEYWORD]


def get_caller_module_locals(depth: int = 1) -> dict[str, Any]:
    """Get the caller module locals dictionary.

    We use sys._getframe instead of inspect.stack(0) because the latter is way slower, since it iterates over
    all the frames in the stack.
    """
    return _getframe(depth).f_locals


def get_caller_module_path(depth: int = 1) -> str:
    """Get the caller module path.

    We use sys._getframe instead of inspect.stack(0) because the latter is way slower, since it iterates over
    all the frames in the stack.
    """
    frame = _getframe(depth)
    return getframeinfo(frame, context=0).filename


_DUMP_START = "_pytest_bdd_>>>"
_DUMP_END = "<<<_pytest_bdd_"


def dump_obj(*objects: Any) -> None:
    """Dump objects to stdout so that they can be inspected by the test suite."""
    for obj in objects:
        dump = pickle.dumps(obj, protocol=pickle.HIGHEST_PROTOCOL)
        encoded = base64.b64encode(dump).decode("ascii")
        print(f"{_DUMP_START}{encoded}{_DUMP_END}")


def collect_dumped_objects(result: RunResult):
    """Parse all the objects dumped with `dump_object` from the result.

    Note: You must run the result with output to stdout enabled.
    For example, using ``testdir.runpytest("-s")``.
    """
    stdout = result.stdout.str()  # pytest < 6.2, otherwise we could just do str(result.stdout)
    payloads = re.findall(rf"{_DUMP_START}(.*?){_DUMP_END}", stdout)
    return [pickle.loads(base64.b64decode(payload)) for payload in payloads]


@attrs
class SimpleMapping(Mapping):
    _dict: dict = attrib(default=Factory(dict), kw_only=True)

    def __getitem__(self, item):
        return self._dict[item]

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)


class DefaultMapping(defaultdict):
    Skip = object()

    def __init__(self, *args, default_factory=None, warm_up_keys=(), **kwargs):
        super().__init__(default_factory, *args, **kwargs)
        self.warm_up(*warm_up_keys)

    def __missing__(self, key):
        if ... in self.keys():
            intercessor = self[...]
            if intercessor is self.Skip:
                raise KeyError(key)
            elif isinstance(intercessor, Callable):
                value = intercessor(key)
            elif intercessor is ...:
                value = key
            else:
                value = intercessor
            self[key] = value
            return value
        else:
            return super().__missing__(key)

    def warm_up(self, *items):
        for item in items:
            with suppress(KeyError):
                getitem(self, item)

    @classmethod
    def instantiate_from_collection_or_bool(
        cls, bool_or_items: Collection[str] | dict[str, Any] | Any = True, *, warm_up_keys=()
    ):
        if isinstance(bool_or_items, Collection):
            if not isinstance(bool_or_items, Mapping):
                bool_or_items = zip(*tee(iter(bool_or_items)))
        else:
            bool_or_items = cast(dict, {...: ...} if bool_or_items else {...: DefaultMapping.Skip})
        return cls(bool_or_items, warm_up_keys=warm_up_keys)


def inject_fixture(request, arg, value):
    """Inject fixture into pytest fixture request.

    :param request: pytest fixture request
    :param arg: argument name
    :param value: argument value
    """

    fd = FixtureDef(
        fixturemanager=request._fixturemanager,
        baseid=None,
        argname=arg,
        func=lambda: value,
        scope="function",
        params=None,
    )
    fd.cached_result = (value, 0, None)

    old_fd = request._fixture_defs.get(arg)
    add_fixturename = arg not in request.fixturenames

    def fin():
        request._fixturemanager._arg2fixturedefs[arg].remove(fd)
        request._fixture_defs[arg] = old_fd

        if add_fixturename:
            request._pyfuncitem._fixtureinfo.names_closure.remove(arg)

    request.addfinalizer(fin)

    # inject fixture definition
    request._fixturemanager._arg2fixturedefs.setdefault(arg, []).insert(0, fd)
    # inject fixture value in request cache
    request._fixture_defs[arg] = fd
    if add_fixturename:
        request._pyfuncitem._fixtureinfo.names_closure.append(arg)


class ModelSchemaPostLoadable:
    postbuild_attrs: list[str] = []

    @staticmethod
    def build_from_schema(cls, data, many, **kwargs):
        return cls.postbuild_attr_builder(cls, data, cls.postbuild_attrs)

    @classmethod
    def schema_post_loader(cls):
        return post_load(partial(cls.build_from_schema, cls))

    @staticmethod
    def postbuild_attr_builder(cls, data, postbuild_args):
        _data = {**data}
        empty = object()
        postbuildable_args = []
        for argument in postbuild_args:
            value = _data.pop(argument, empty)
            if value is not empty:
                postbuildable_args.append((argument, value))
        instance = cls(**_data)
        for argument, value in postbuildable_args:
            setattr(instance, argument, value)
        return instance

    def setattrs(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)
        return self


def _itemgetter(*items):
    def func(obj):
        if len(items) == 0:
            return []
        elif len(items) == 1:
            return [obj[items[0]]]
        else:
            return itemgetter(*items)(obj)

    return func


class _NoneException(Exception):
    ...


def deepattrgetter(*attrs, **kwargs):
    empty = object()
    default = kwargs.pop("default", empty)
    default_exception_type = AttributeError if default is not empty else _NoneException
    skip_missing = kwargs.pop("skip_missing", False)
    skip_missing_context = suppress(AttributeError) if skip_missing else nullcontext()
    if default is not empty and skip_missing:
        raise ValueError('Both "default" and "skip_missing" are specified')

    def fn(obj):
        def _():
            for attr in attrs:
                try:
                    with skip_missing_context:
                        yield attrgetter(attr)(obj)
                except default_exception_type:
                    yield default

        return tuple(_())

    return fn


def make_python_name(string: str) -> str:
    """Make python attribute name out of a given string."""
    string = re.sub(PYTHON_REPLACE_REGEX, "", string.replace(" ", "_"))
    return re.sub(ALPHA_REGEX, "", string).lower()



# Refactor to Enums
import re
from collections import defaultdict

TAG = "tag"
FEATURE = "feature"
SCENARIO_OUTLINE = "scenario outline"
EXAMPLES = "examples"
EXAMPLES_VERTICAL = "examples vertical"
SCENARIO = "scenario"
BACKGROUND = "background"
EXAMPLES_HEADERS = "example headers"
EXAMPLE_LINE = "example line"
EXAMPLE_LINE_VERTICAL = "example line vertical"


class STEP_TYPE:
    GIVEN = "given"
    WHEN = "when"
    THEN = "then"
    AND = "and"
    OTHER = "other"


STEP_PREFIXES = {
    FEATURE: "Feature: ",
    SCENARIO_OUTLINE: "Scenario Outline: ",
    EXAMPLES_VERTICAL: "Examples: Vertical",
    EXAMPLES: "Examples:",
    SCENARIO: "Scenario: ",
    BACKGROUND: "Background:",
    STEP_TYPE.GIVEN: "Given ",
    STEP_TYPE.WHEN: "When ",
    STEP_TYPE.THEN: "Then ",
    TAG: "@",
}
STEP_TYPES_BY_NORMALIZED_PREFIX = defaultdict(
    lambda: STEP_TYPE.OTHER,
    {
        "given": STEP_TYPE.GIVEN,
        "when": STEP_TYPE.WHEN,
        "then": STEP_TYPE.THEN,
        "and": STEP_TYPE.AND,
        "but": STEP_TYPE.AND,
        "*": STEP_TYPE.AND,
    },
)
PYTHON_REPLACE_REGEX = re.compile(r"\W")
ALPHA_REGEX = re.compile(r"^\d+_*")



"""pytest-bdd public API."""
from __future__ import annotations

from pytest_bdd.packaging import get_distribution_version
from pytest_bdd.scenario import scenario, scenarios
from pytest_bdd.steps import given, step, then, when
from pytest_bdd.warning_types import (
    PytestBDDScenarioExamplesExtraParamsWarning,
    PytestBDDScenarioStepsExtraPramsWarning,
    PytestBDDStepDefinitionWarning,
)

__version__ = str(get_distribution_version("pytest-bdd-ng"))

__all__ = ["given", "when", "then", "scenario", "scenarios"]



from __future__ import annotations

from itertools import count
from operator import attrgetter
from typing import Any

from attr import Factory, attrib, attrs

import pytest_bdd.ast as ast
from pytest_bdd.struct_bdd.model import Join, Step, Table


@attrs
class _ASTBuilder:
    model: Any
    id_generator = attrib(default=Factory(lambda: map(str, count())), kw_only=True)

    def build(self, *args, **kwargs):  # pragma: no cover
        raise NotImplementedError


@attrs
class DocumentASTBuilder(_ASTBuilder):
    model: Step = attrib()

    def build(self):
        return ast.AST(
            gherkin_document=ast.GherkinDocument(comments=[], uri=None).setattrs(
                feature=StepToFeatureASTBuilder(self.model, id_generator=self.id_generator).build()
            )
        )


@attrs
class StepToFeatureASTBuilder(_ASTBuilder):
    model: Step = attrib()

    def build(self):
        return ast.Feature(
            children=self._build_children(),
            description=self.model.description or "",
            language="EN",
            location=ast.Location(column=0, line=0),
            tags=[],
            name=self.model.name or "",
        ).setattrs(keyword="Feature")

    def _build_children(self):
        def _():
            for route in self.model.routes:
                if route.steps:
                    yield ast.NodeContainerChild().setattrs(
                        scenario=ast.Scenario(
                            description=route.steps[0].description or "",
                            examples=[ExampleASTBuilder(route.example_table, id_generator=self.id_generator).build()]
                            if route.example_table.values
                            else [],
                            identifier=next(self.id_generator),
                            keyword="Scenario",
                            location=ast.Location(column=0, line=0),
                            name=next(filter(bool, map(attrgetter("name"), reversed(route.steps))), ""),
                            tags=[
                                *map(
                                    lambda tag_name: ast.Tag(
                                        identifier=next(self.id_generator),
                                        location=ast.Location(column=0, line=0),
                                        name=tag_name,
                                    ),
                                    route.tags,
                                )
                            ],
                            steps=[
                                *map(
                                    lambda step: ast.Step(
                                        identifier=next(self.id_generator),
                                        keyword=step.type or "Given",
                                        location=ast.Location(column=0, line=0),
                                        text=step.action,
                                    ).setattrs(
                                        data_table=ast.DataTable(
                                            rows=[
                                                *map(
                                                    lambda row_values: ast.TableRow(
                                                        identifier=next(self.id_generator),
                                                        location=ast.Location(column=0, line=0),
                                                        cells=[
                                                            *map(
                                                                lambda parameter: ast.TableCell(
                                                                    location=ast.Location(column=0, line=0),
                                                                    value=parameter,
                                                                ),
                                                                row_values,
                                                            )
                                                        ],
                                                    ),
                                                    Join(tables=step.data).rowed_values,
                                                )
                                            ],
                                            location=ast.Location(column=0, line=0),
                                        ),
                                        **(
                                            {
                                                "doc_string": ast.DocString(
                                                    content=step.description,
                                                    delimiter="\n",
                                                    location=ast.Location(column=0, line=0),
                                                )
                                            }
                                            if step.description
                                            else {}
                                        ),
                                    ),
                                    filter(lambda step: step.action is not None, route.steps),
                                )
                            ],
                        )
                    )

        return list(_())


@attrs
class ExampleASTBuilder(_ASTBuilder):
    model: Table | Join = attrib()

    def build(self):
        return ast.Example(
            description=self.model.description,
            identifier=next(self.id_generator),
            keyword="Examples",
            location=ast.Location(column=0, line=0),
            name=self.model.name,
            table_body=[
                *map(
                    lambda row_values: ast.TableRow(
                        identifier=next(self.id_generator),
                        location=ast.Location(column=0, line=0),
                        cells=[
                            *map(
                                lambda parameter: ast.TableCell(
                                    location=ast.Location(column=0, line=0),
                                    value=parameter,
                                ),
                                row_values,
                            )
                        ],
                    ),
                    self.model.rowed_values,
                )
            ],
            tags=[
                *map(
                    lambda tag_name: ast.Tag(
                        identifier=next(self.id_generator),
                        location=ast.Location(column=0, line=0),
                        name=tag_name,
                    ),
                    self.model.tags,
                )
            ],
        ).setattrs(
            table_header=ast.ExampleTableHeader(
                identifier=next(self.id_generator),
                location=ast.Location(column=0, line=0),
                cells=[
                    *map(
                        lambda parameter: ast.TableCell(
                            location=ast.Location(column=0, line=0),
                            value=parameter,
                        ),
                        self.model.parameters,
                    )
                ],
            ),
        )



from functools import partial
from operator import methodcaller
from pathlib import Path
from typing import cast

from attr import attrib, attrs

from pytest_bdd.struct_bdd.model import Step, StepSchema
from pytest_bdd.typing.parser import ParserProtocol


@attrs
class StructBDDParser(ParserProtocol):
    class KIND:
        HOCON = "hocon"
        HJSON = "hjson"
        JSON = "json"
        JSON5 = "json5"
        TOML = "toml"
        YAML = "yaml"

    kind = attrib(kw_only=True)
    glob = attrib(kw_only=True)
    loader = attrib(kw_only=True)

    @kind.default
    def kind_default(self):
        return self.KIND.YAML if getattr(self, "loader", None) is None else None

    @glob.default
    def glob_default(self):
        return methodcaller("rglob", "*" if self.kind is None else f"*.bdd.{self.kind}")

    @loader.default
    def loader_default(self):
        return self.build_loader()

    def parse(self, path: Path, uri: str, *args, **kwargs):
        encoding = kwargs.pop("encoding", "utf-8")
        mode = kwargs.pop("mode", "r")
        with path.open(mode=mode, encoding=encoding) as feature_file:
            content = feature_file.read()

        return cast(Step, StepSchema().load(self.loader(content, *args, **kwargs))).build_feature(
            filename=str(path.as_posix()), uri=uri
        )

    def build_loader(self):
        if self.kind == self.KIND.YAML:
            from yaml import FullLoader
            from yaml import load as load_yaml

            return partial(load_yaml, Loader=FullLoader)
        elif self.kind == self.KIND.TOML:
            from tomli import loads as load_toml

            return load_toml
        elif self.kind == self.KIND.JSON:
            from json import loads as load_json

            return load_json
        elif self.kind == self.KIND.JSON5:
            from json5 import loads as load_json5

            return load_json5
        elif self.kind == self.KIND.HJSON:
            from hjson import loads as load_hjson

            return load_hjson
        elif self.kind == self.KIND.HOCON:
            from json import loads

            from pyhocon import ConfigFactory, HOCONConverter

            def load_hocon(
                s,
                hocon_parse_args=(),
                hocon_parse_kwargs=None,
                hocon_to_json_args=(),
                hocon_to_json_kwargs=None,
                json_args=(),
                json_kwargs=None,
            ):
                hocon_to_json_kwargs = hocon_to_json_kwargs or {}
                hocon_parse_kwargs = hocon_parse_kwargs or {}
                json_kwargs = json_kwargs or {}
                return loads(
                    HOCONConverter.to_json(
                        config=ConfigFactory.parse_string(s, *hocon_parse_args, **hocon_parse_kwargs),
                        *hocon_to_json_args,
                        **hocon_to_json_kwargs,
                    ),
                    *json_args,
                    **json_kwargs,
                )

            return load_hocon



"""Pytest plugin entry point. Used for any fixtures needed."""
from __future__ import annotations

from collections import deque
from contextlib import suppress
from typing import Collection

import pytest

from pytest_bdd import cucumber_json, generation, gherkin_terminal_reporter, given, steps, then, when
from pytest_bdd.allure_logging import AllurePytestBDD
from pytest_bdd.model import Step
from pytest_bdd.reporting import ScenarioReporterPlugin
from pytest_bdd.runner import ScenarioRunner
from pytest_bdd.steps import StepHandler
from pytest_bdd.typing.pytest import Config, Mark, MarkDecorator, Metafunc, Parser, PytestPluginManager


def pytest_addhooks(pluginmanager: PytestPluginManager) -> None:
    """Register plugin hooks."""
    from pytest_bdd import hooks

    pluginmanager.add_hookspecs(hooks)


@given("trace")
@when("trace")
@then("trace")
def trace() -> None:
    """Enter pytest's pdb trace."""
    pytest.set_trace()


@pytest.fixture
def step_registry() -> StepHandler.Registry:
    """Fixture containing registry of all user-defined steps"""


@pytest.fixture
def step_matcher(pytestconfig) -> StepHandler.Matcher:
    """Fixture containing matcher to help find step definition for selected step of scenario"""
    return StepHandler.Matcher(pytestconfig)  # type: ignore[call-arg]


@pytest.fixture
def steps_left() -> deque[Step]:
    """Fixture containing steps which are left to be executed"""
    return deque()


def pytest_addoption(parser: Parser) -> None:
    """Add pytest-bdd options."""
    add_bdd_ini(parser)
    steps.add_options(parser)
    cucumber_json.add_options(parser)
    generation.add_options(parser)
    gherkin_terminal_reporter.add_options(parser)


def add_bdd_ini(parser: Parser) -> None:
    parser.addini("bdd_features_base_dir", "Base features directory.")


@pytest.mark.trylast
def pytest_configure(config: Config) -> None:
    """Configure all subplugins."""
    config.addinivalue_line("markers", "pytest_bdd_scenario: marker to identify pytest_bdd tests")
    cucumber_json.configure(config)
    gherkin_terminal_reporter.configure(config)
    config.pluginmanager.register(ScenarioReporterPlugin())
    config.pluginmanager.register(ScenarioRunner())
    AllurePytestBDD.register_if_allure_accessible(config)


def pytest_unconfigure(config: Config) -> None:
    """Unconfigure all subplugins."""
    cucumber_json.unconfigure(config)


def pytest_generate_tests(metafunc: Metafunc):
    with suppress(AttributeError):
        metafunc.function.__pytest_bdd_scenario_registry__.parametrize(metafunc)


def pytest_cmdline_main(config: Config) -> int | None:
    return generation.cmdline_main(config)


@pytest.mark.trylast
def pytest_bdd_convert_tag_to_marks(feature, scenario, tag) -> Collection[Mark | MarkDecorator] | None:
    return [getattr(pytest.mark, tag)]


def pytest_bdd_match_step_definition_to_step(request, feature, scenario, step, previous_step) -> StepHandler.Definition:
    step_registry: StepHandler.Registry = request.getfixturevalue("step_registry")
    step_matcher: StepHandler.Matcher = request.getfixturevalue("step_matcher")

    return step_matcher(feature, scenario, step, previous_step, step_registry)



"""Common type definitions."""
from __future__ import annotations

import sys

if sys.version_info >= (3, 8):
    from typing import Protocol, runtime_checkable
else:
    from typing_extensions import Protocol, runtime_checkable

assert Protocol
assert runtime_checkable
